/*****************************************************************************/
#ifdef COMMENTS_WITH_COMMENTS
/*
                                  DCLinabox.c

JavaScript based VT100 terminal emulation using WASD WebSocket communication
with a Pseudo-terminal.

Based on JavaScript code from the ShellInABox project:

  Copyright (C) 2008-2009 Markus Gutschke <markus@shellinabox.com>.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

http://code.google.com/p/shellinabox/

For all things VT100 (et.al.) ... http://vt100.net


CAUTION
-------
This WebSocket application allows remote login from a Web browser to the server
system.  This could be a security issue and so the script disables itself by
default.  Logical name value DCLINABOX_ENABLE controls whether this script can
be used.   Define this system-wide using a value of "*" to simply allow
experimentation.  Alternatively provide one or more comma-separated,
dotted-decimal IP address to specify one or more hosts allowed to use the
script, and/or one or more comman-separated IP addresses and CIDR subnet mask
to specify a range of hosts.  IPv4 only!  For example

  $ DEFINE /SYSTEM DCLINABOX_ENABLE "*"
  $ DEFINE /SYSTEM DCLINABOX_ENABLE "192.168.1.2"
  $ DEFINE /SYSTEM DCLINABOX_ENABLE "192.168.1.2,192.168.1.3"
  $ DEFINE /SYSTEM DCLINABOX_ENABLE "192.168.1.0/24"
  $ DEFINE /SYSTEM DCLINABOX_ENABLE "192.168.1.0/24,192.168.2.2"

By default the WebSocket, and hence all traffic to and from the DCLinabox login
and session, is only allowed over Secure Sockets Layer.  To allow access via
clear-text connections add "ws:" somewhere in the logical name value.

  $ DEFINE /SYSTEM DCLINABOX_ENABLE "ws:,*"
  $ DEFINE /SYSTEM DCLINABOX_ENABLE "192.168.1.0/24,ws:,192.168.2.2"

The logical name DCLINABOX_IDLE allows the number of minutes before client
disconnection to be specified.  Define to -1 to to disable idle diconnection.

  $ DEFINE /SYSTEM DCLINABOX_IDLE 1440
  $ DEFINE /SYSTEM DCLINABOX_IDLE -1


RENAMING THE INTERFACE
-----------------------
The DCLINABOX.EXE file may be alternately named.  With a public system this may
be useful for reducing nuisance-value access attempts and/or an obvious attack
vector by obscuring the web interface.  Just use an obscured (or
access-controlled) HTML terminal page, a different script executable file name,
and add that script name to the <head>..</head> section of the HTML terminal
file as with other per-terminal configuration variables.

  <script type="text/javascript"><!--
  DCLinaboxScriptName = "anexample";
  //--></script>

The script accessed will then be /cgiplus-bin/anexample.

The logical names used would then be ANEXAMPLE_ENABLE, etc.


COPYRIGHT
---------
Copyright (C) 2011,2012 Mark G.Daniel
This program comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY
---------------
21-JUL-2012  MGD  v1.0.1, bugfix; ptd$delete() during client removal
                          bugfix; AddClient() lib$free_vm_page() memory
04-DEC-2011  MGD  v1.0.0, initial development
*/
#endif /* COMMENTS_WITH_COMMENTS */
/*****************************************************************************/

#define SOFTWAREVN "1.0.1"
#define SOFTWARENM "DCLINABOX"
#ifdef __ALPHA
#  define SOFTWAREID SOFTWARENM " AXP-" SOFTWAREVN
#endif
#ifdef __ia64
#  define SOFTWAREID SOFTWARENM " IA64-" SOFTWAREVN
#endif
#ifdef __VAX
#  define SOFTWAREID SOFTWARENM " VAX-" SOFTWAREVN
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>

#include <descrip.h>
#include <iodef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>
#include <ttdef.h>
#include <tt2def.h>

#include "wslib.h"

#define DC$_TERM 6

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) !(((x) & STS$M_SUCCESS))

#define Debug 0

#define FI_LI "DCLINABOX", __LINE__
#define EXIT_FI_LI(status) { printf ("[%s:%d]", FI_LI); exit(status); }

/* basically the page size on the architecture */
#ifdef VAX
#define PTD_READ_SIZE 512
#define PTD_WRITE_SIZE 512
#else
#define PTD_READ_SIZE 8192
#define PTD_WRITE_SIZE 8192
#endif

int  ConnectedCount,
     UsageCount;

char  EnableLogicalName [128],
      IdleLogicalName [128];

struct PtdClient {

   /* keep these adjacent and aligned on a page boundary */
   char  PtdReadBuffer [PTD_READ_SIZE],
         PtdWriteBuffer [PTD_WRITE_SIZE];

   int  PtdQueuedRead,
        PtdQueuedWrite,
        PtdReadCount,
        PtdWriteCount;

   unsigned short  ptdchan;

   char  InputBuffer [256];

   struct WsLibStruct  *WsLibPtr;
};

long  PtdClientPages = (sizeof(struct PtdClient) / 512 ) + 1;

/* function prototypes */
void AddClient ();
int DCLinaboxEnable ();
void ReadClient (struct WsLibStruct*);
void RemoveClient (struct WsLibStruct *wsptr);
void RemoteHostPort (struct PtdClient*);

int PtdOpen (struct PtdClient*);
void PtdClose (struct PtdClient*);
void PtdRead (struct PtdClient*);
void PtdReadAst (struct PtdClient*);
void PtdReadWriteAst (struct WsLibStruct*);
void PtdWrite (struct PtdClient*, char*, int);
void PtdWriteAst (struct PtdClient*);

/*****************************************************************************/
/*
AST delivery is disabled during client acceptance and the add-client function
is deferred using an AST to help minimise the client setup window with a
potentially busy WebSocket application.
*/

main (int argc, char *argv[])

{
   char  *aptr, *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (argc > 1 && !strcasecmp(argv[1],"/VERSION"))
   {
      fprintf (stdout, "%%DCLINABOX-I-VERSION, %s with %s\n",
               SOFTWAREID, WsLibVersion());
      exit (SS$_NORMAL);
   }

   /* don't want the C-RTL fiddling with the carriage control */
   stdout = freopen ("SYS$OUTPUT", "w", stdout, "ctx=bin");

   for (aptr = argv[0]; *aptr; aptr++);
   while (aptr > argv[0] && *aptr != ']') aptr--;
   if (*aptr++ != ']') exit (SS$_BUGCHECK);

   /* generate the ENABLE logical name from the executable file name */
   zptr = (sptr = EnableLogicalName) + sizeof(EnableLogicalName)-1;
   for (cptr = aptr;
        *cptr && *cptr != '.' && sptr < zptr;
        *sptr++ = toupper(*cptr++));
   for (cptr = "_ENABLE"; *cptr && sptr < zptr; *sptr++ = *cptr++);
   *sptr = '\0';

   /* generate the IDLE logical name from the executable file name */
   zptr = (sptr = IdleLogicalName) + sizeof(IdleLogicalName)-1;
   for (cptr = aptr;
        *cptr && *cptr != '.' && sptr < zptr;
        *sptr++ = toupper(*cptr++));
   for (cptr = "_IDLE"; *cptr && sptr < zptr; *sptr++ = *cptr++);
   *sptr = '\0';

   /* no clients is two minutes in seconds */
   WsLibSetLifeSecs (2*60);

   while (WsLibIsCgiPlus())
   {
      WsLibCgiVar ("");

      sys$setast (0);

      UsageCount++;

      if (DCLinaboxEnable()) AddClient ();

      WsLibCgiPlusEof ();

      sys$setast (1);
   }

   exit (SS$_NORMAL);
}

/*****************************************************************************/
/*
Allocate a client structure and add it to the head of the list.
*/

void AddClient ()

{
   int  len, status,
        IdleMins = 0;
   char  *cptr;
   char  ErrorMsg [256];
   struct PtdClient  *clptr;

   /*********/
   /* begin */
   /*********/

   /* PTD$ buffers must be page aligned */
   status = lib$get_vm_page (&PtdClientPages, &clptr);
   if (VMSnok(status)) EXIT_FI_LI (status);
   memset (clptr, 0, sizeof(struct PtdClient));

   /* create a WebSocket library structure for the client */
   if (!(clptr->WsLibPtr = WsLibCreate (clptr, RemoveClient)))
   {
      /* failed, commonly on some WebSocket protocol issue */
      status = lib$free_vm_page (&PtdClientPages, &clptr);
      if (VMSnok(status)) EXIT_FI_LI (status);
      return;
   }

   if (cptr = getenv (IdleLogicalName)) IdleMins = atoi(cptr);
   if (!IdleMins) IdleMins = 120;
   if (IdleMins < 0) IdleMins = 365*24*60;
   WsLibSetIdleSecs (clptr->WsLibPtr, IdleMins*60);

   /* open the IPC to the WebSocket (mailboxes) */
   status = WsLibOpen (clptr->WsLibPtr);
   if (VMSnok(status)) EXIT_FI_LI(status);

   status = PtdOpen (clptr);
   if (VMSnok (status))
   {
      len = sprintf (ErrorMsg, "%%X%08.08X", status);
      WsLibWrite (clptr->WsLibPtr, ErrorMsg, len, WSLIB_ASYNCH);
      WsLibClose (clptr->WsLibPtr, 0, NULL);
      return;
   }

   /* queue an asynchronous read from the client */
   WsLibRead (clptr->WsLibPtr,
              clptr->InputBuffer,
              sizeof(clptr->InputBuffer),
              ReadClient);

   ConnectedCount++;
}

/*****************************************************************************/
/*
Remove the client structure from the list and free the memory.
*/

void RemoveClient (struct WsLibStruct *wsptr)

{
   int  status;
   struct PtdClient  *clptr;

   /*********/
   /* begin */
   /*********/

   clptr = WsLibGetUserData (wsptr);

   if (clptr->ptdchan) status = ptd$delete (clptr->ptdchan);

   status = lib$free_vm_page (&PtdClientPages, &clptr);
   if (VMSnok(status)) EXIT_FI_LI (status);

   if (ConnectedCount) ConnectedCount--;
}

/*****************************************************************************/
/*
*/

int PtdOpen (struct PtdClient *clptr)

{
   int  status;
   long  InAdr [2],
         CharBuf [3];

   /*********/
   /* begin */
   /*********/

   CharBuf[0] = (80 << 16) | (TT$_LA100 << 8) | DC$_TERM;
   CharBuf[1] = (24 << 24) |
                TT$M_EIGHTBIT | TT$M_SCOPE | TT$M_WRAP |
                TT$M_MECHTAB | TT$M_LOWER | TT$M_TTSYNC;
   CharBuf[2] = TT2$M_EDIT | TT2$M_DRCS | TT2$M_EDITING | TT2$M_HANGUP;

   InAdr[0] = (int)(clptr->PtdReadBuffer);
   InAdr[1] = (int)(clptr->PtdReadBuffer +
                    sizeof(clptr->PtdReadBuffer) +
                    sizeof(clptr->PtdWriteBuffer)-1);

   status = ptd$create (&clptr->ptdchan, 0, CharBuf, sizeof(CharBuf),
                        0, 0, 0, InAdr);
   if (VMSnok (status)) return (status);

   clptr->PtdQueuedRead++;
   status = ptd$read (0, clptr->ptdchan, &PtdReadAst, clptr,
                      clptr->PtdReadBuffer, sizeof(clptr->PtdReadBuffer));

   return (status);
}

/*****************************************************************************/
/*
*/

void PtdClose (struct PtdClient *clptr)

{
   /*********/
   /* begin */
   /*********/

   if (clptr->PtdQueuedRead || clptr->PtdQueuedRead)
   {
      ptd$cancel (clptr->ptdchan);
      return;
   }

   WsLibClose (clptr->WsLibPtr, 0 , NULL);
}

/*****************************************************************************/
/*
Data has been read from the PTD (i.e. from the system).
*/

void PtdReadAst (struct PtdClient *clptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (clptr->PtdQueuedRead) clptr->PtdQueuedRead--;

   status = *(short*)clptr->PtdReadBuffer;
   if (VMSok(status))
      WsLibWrite (clptr->WsLibPtr,
                  clptr->PtdReadBuffer + sizeof(short)+sizeof(short),
                  *(short*)(clptr->PtdReadBuffer + sizeof(short)),
                  PtdReadWriteAst);
   else
      PtdClose (clptr);
}

/*****************************************************************************/
/*
Data read from the PTD (system) has been written to the WebSocket client. 
Check status and if OK queue another read from the PTD.
*/

void PtdReadWriteAst (struct WsLibStruct *wsptr)

{
   int  status;
   struct PtdClient  *clptr;

   /*********/
   /* begin */
   /*********/

   clptr = WsLibGetUserData(wsptr);

   status = WsLibWriteStatus (wsptr);
   if (VMSok (status))
   {
      clptr->PtdQueuedRead++;
      ptd$read (0, clptr->ptdchan, &PtdReadAst, clptr,
                clptr->PtdReadBuffer, sizeof(clptr->PtdReadBuffer));
   }
   else
      WsLibClose (wsptr, 0, NULL);
}

/*****************************************************************************/
/*
Asynchronous read from a WebSocket client has concluded.
*/

void ReadClient (struct WsLibStruct *wsptr)

{
   int  cnt;
   struct PtdClient  *clptr;

   /*********/
   /* begin */
   /*********/

   if (VMSnok (WsLibReadStatus(wsptr)))
   {
      /* WEBSOCKET_INPUT read error (can be EOF) */
      WsLibClose (wsptr, 0, NULL);
      return;
   }

   clptr = WsLibGetUserData(wsptr);

   if (cnt = WsLibReadCount(wsptr))
      PtdWrite (clptr, clptr->InputBuffer, cnt);
   else
      /* otherwise queue the next asynchronous read from the client */
      WsLibRead (wsptr,
                 clptr->InputBuffer,
                 sizeof(clptr->InputBuffer),
                 ReadClient);
}

/*****************************************************************************/
/*
Write the supplied data to the PTD (i.e. to the system).
*/

void PtdWrite
(
struct PtdClient *clptr,
char *DataPtr,
int DataCount
)
{
   int  status;
   char  *bptr, *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   sptr = bptr = clptr->PtdWriteBuffer + sizeof(short)+sizeof(short);
   zptr = sptr + sizeof(clptr->PtdWriteBuffer) - sizeof(short)-sizeof(short);
   if (DataCount < 0) DataCount = strlen(DataPtr);
   for (cptr = DataPtr; DataCount-- && sptr < zptr; *sptr++ = *cptr++);
   clptr->PtdWriteCount = sptr - bptr;

   clptr->PtdQueuedWrite++;
   ptd$write (clptr->ptdchan, PtdWriteAst, clptr,
              clptr->PtdWriteBuffer, clptr->PtdWriteCount, 0, 0);
}

/*****************************************************************************/
/*
PTD write (to system) has completed.  If OK read from the WebSocket client.
*/

void PtdWriteAst (struct PtdClient *clptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (clptr->PtdQueuedWrite) clptr->PtdQueuedWrite--;

   status = *(short*)clptr->PtdWriteBuffer;
   if (VMSok(status) ||
       status == SS$_DATAOVERUN ||
       status == SS$_DATALOST)
      WsLibRead (clptr->WsLibPtr,
                 clptr->InputBuffer,
                 sizeof(clptr->InputBuffer),
                 ReadClient);
   else
      PtdClose (clptr);
}

/*****************************************************************************/
/*
Logical name value DCLINABOX_ENABLE controls whether this script can be used. 
Make the value "*" to allow all remote hosts.  Alternatively provide one or
more comma-separated, dotted-decimal IP address to specify one or more hosts
allowed to use the script, and/or one or more comman-separated IP addresses and
CIDR subnet mask to specify a range of hosts.  IPv4 only!
*/

int DCLinaboxEnable ()

{
   unsigned int  IPaddr, IPnet, mask;
   char  ch;
   char  *aptr, *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (!(cptr = getenv (EnableLogicalName)))
   {
      fprintf (stdout, "Status: 403 \"%s\" undefined\r\n\r\n",
               EnableLogicalName);
      exit (1);
   }

   if (!(aptr = WsLibCgiVar("REMOTE_ADDR"))) return (0);
   /* inet_aton() is not available on VMS V7.2 */
   IPaddr = inet_addr (aptr);
   if (IPaddr == 0xffffffff) return (0);

   if (!strstr (cptr, "ws:"))
   {
      if (!(sptr = WsLibCgiVar ("REQUEST_SCHEME"))) return (0);
      if (strcmp(sptr,"wss:") && strcmp(sptr,"https:"))
      {
         fprintf (stdout, "Status: 403 Must be SSL\r\n\r\n");
         return (0);
      }
   }

   if (strchr (cptr, '*')) return (1);

   while (*cptr && *sptr)
   {
      while (*cptr && *cptr != ',') cptr++;
      if (ch = *cptr) *cptr = '\0';
      if (zptr = strchr (sptr, '/'))
      {
         /* subnet mask */
         *zptr = '\0';
         /* inet_aton() is not available on VMS V7.2 */
         IPnet = inet_addr (sptr);
         if (IPnet == 0xffffffff) return (0);
         mask = atoi(zptr+1);
         mask = 0xffffffff >> (32 - mask); 
         /* if both addresses are valid and masked client address matches */
         if (IPaddr && IPnet && (IPnet == (IPaddr & mask))) return (1);
         *zptr = '/';
      }
      else
      if (!strcmp (sptr, aptr))
         return (1);
      if (*cptr = ch) cptr++;
      sptr = cptr;
   }

   fprintf (stdout, "Status: 403 Not Permitted\r\n\r\n");

   return (0);
}

/*****************************************************************************/

